;; xgen.scm extracted definitions from /usr/include/X11/Xatom.h
(define PRIMARY 1)
(define SECONDARY 2)
(define ARC 3)
(define ATOM 4)
(define BITMAP 5)
(define CARDINAL 6)
(define COLORMAP 7)
(define CURSOR 8)
(define CUT-BUFFER0 9)
(define CUT-BUFFER1 10)
(define CUT-BUFFER2 11)
(define CUT-BUFFER3 12)
(define CUT-BUFFER4 13)
(define CUT-BUFFER5 14)
(define CUT-BUFFER6 15)
(define CUT-BUFFER7 16)
(define DRAWABLE 17)
(define FONT 18)
(define INTEGER 19)
(define PIXMAP 20)
(define POINT 21)
(define RECTANGLE 22)
(define RESOURCE-MANAGER 23)
(define RGB-COLOR-MAP 24)
(define RGB-BEST-MAP 25)
(define RGB-BLUE-MAP 26)
(define RGB-DEFAULT-MAP 27)
(define RGB-GRAY-MAP 28)
(define RGB-GREEN-MAP 29)
(define RGB-RED-MAP 30)
(define STRING 31)
(define VISUALID 32)
(define WINDOW 33)
(define WM-COMMAND 34)
(define WM-HINTS 35)
(define WM-CLIENT-MACHINE 36)
(define WM-ICON-NAME 37)
(define WM-ICON-SIZE 38)
(define WM-NAME 39)
(define WM-NORMAL-HINTS 40)
(define WM-SIZE-HINTS 41)
(define WM-ZOOM-HINTS 42)
(define MIN-SPACE 43)
(define NORM-SPACE 44)
(define MAX-SPACE 45)
(define END-SPACE 46)
(define SUPERSCRIPT-X 47)
(define SUPERSCRIPT-Y 48)
(define SUBSCRIPT-X 49)
(define SUBSCRIPT-Y 50)
(define UNDERLINE-POSITION 51)
(define UNDERLINE-THICKNESS 52)
(define STRIKEOUT-ASCENT 53)
(define STRIKEOUT-DESCENT 54)
(define ITALIC-ANGLE 55)
(define X-HEIGHT 56)
(define QUAD-WIDTH 57)
(define WEIGHT 58)
(define POINT-SIZE 59)
(define RESOLUTION 60)
(define COPYRIGHT 61)
(define NOTICE 62)
(define FONT-NAME 63)
(define FAMILY-NAME 64)
(define FULL-NAME 65)
(define CAP-HEIGHT 66)
(define WM-CLASS 67)
(define WM-TRANSIENT-FOR 68)
(define LAST-PREDEFINED 68)
;; xgen.scm extracted definitions from /usr/include/X11/Xcms.h
(define	X:Undefined	#x00000000)
(define	X:CIEXYZ	#x00000001)
(define	X:CIEuvY	#x00000002)
(define	X:CIExyY	#x00000003)
(define	X:CIELab	#x00000004)
(define	X:CIELuv	#x00000005)
(define	X:TekHVC	#x00000006)
(define	X:RGB	#x80000000)
(define	X:RGBi	#x80000001)
