;; xgen.scm extracted typedef structs from /usr/include/X11/Xlib.h
(define X-event:type #x10)
(define X-event:serial #x11)
(define X-event:send-event #x12)
(define X-event:time #x13)
(define X-event:x #x14)
(define X-event:y #x15)
(define X-event:x-root #x16)
(define X-event:y-root #x17)
(define X-event:state #x18)
(define X-event:keycode #x19)
(define X-event:same-screen #x1a)
(define X-event:button #x1b)
(define X-event:is-hint #x1c)
(define X-event:mode #x1d)
(define X-event:detail #x1e)
(define X-event:focus #x1f)
(define X-event:width #x20)
(define X-event:height #x21)
(define X-event:count #x22)
(define X-event:major-code #x23)
(define X-event:minor-code #x24)
(define X-event:border-width #x25)
(define X-event:override-redirect #x26)
(define X-event:from-configure #x27)
(define X-event:value-mask #x28)
(define X-event:place #x29)
(define X-event:new #x2a)
(define X-event:format #x2b)
(define X-event:request #x2c)
(define X-event:first-keycode #x2d)
